/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include <openssl/sha.h>
#include <stdio.h>
#include <string.h>

#include "hash_draft.h"

void
msg_hash256(unsigned char *out, const unsigned char *in,
    const unsigned int inlen)
{
	SHA256(in, inlen, out);
}

void
msg_hash512(unsigned char *out, const unsigned char *in,
    const unsigned int inlen)
{
	SHA512(in, inlen, out);
}

int
F(unsigned char *out, const unsigned char *in, const unsigned char *key,
    const unsigned int len)
{
	/*
	 * returns 1 if the message digest length doesn't fit 32 or 64 bytes
	 * m = n; len is of length n; in is of length 2 * n
	 */
	unsigned char byte[len], left[3 * len], right[2 * len];

	/* replaces toByte for byte */
	memset(byte, 0, sizeof(byte));
	memset(left, 0, sizeof(left));
	memset(right, 0, sizeof(right));
	concat_bytes(right, key, len, in, len);
	concat_bytes(left, byte, len, right, 2 * len);
	if (len == 32) {
		/* toByte(0,32) || key || M */
		msg_hash256(out, left, 3 * len);
	} else if (len == 64) {
		/* toByte(0,64) || key || M */
		msg_hash512(out, left, 3 * len);
	} else {
		return 1;
	}

	return 0;
}

int
H(unsigned char *out, const unsigned char *in, const unsigned char *key,
    const unsigned int len)
{
	/*
	 * returns 1 if the message digest length doesn't fit 32 or 64 bytes
	 * hash(out, in, key, len, 0);
	 * m = n; len is of length n; in is of length 2 * n
	 */
	unsigned char byte[len], left[4 * len], right[3 * len];

	/* replaces toByte for byte */
	memset(byte, 0, sizeof(byte));
	memset(left, 0, sizeof(left));
	memset(right, 0, sizeof(right));
	concat_bytes(right, key, len, in, 2 * len);
	concat_bytes(left, byte, len, right, 3 * len);
	if (len == 32) {
		/* toByte(0,32) || key || M */
		msg_hash256(out, left, 4 * len);
	} else if (len == 64) {
		/* toByte(0,64) || key || M */
		msg_hash512(out, left, 4 * len);
	}
	else {
		return 1;
	}

	return 0;
}

int
Hm(unsigned char *out, const unsigned char *in, const unsigned int msglen,
    const unsigned char *key, const unsigned int len)
{
	/* returns 1 if the message digest length doesn't fit 32 or 64 bytes */
	unsigned char tmp[2 * len + msglen];
	memset(tmp, 0, sizeof(tmp));
	concat_bytes(tmp, key, 2*len, in, msglen);
	if (len == 32) {
		/* KEY || M */
		msg_hash256(out, tmp, 2 * len + msglen);
	} else if (len == 64) {
		/* KEY || M */
		msg_hash512(out, tmp, 2 * len + msglen);
	} else {
		return 1;
	}

	return 0;
}

void
concat_bytes(unsigned char *concat, const unsigned char *left,
    const unsigned int bytes_left, const unsigned char *right,
    const unsigned int bytes_right)
{
	memmove(concat, left, bytes_left);
	memmove(concat + bytes_left, right, bytes_right);
}
