/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "wots_draft.h"

/* May be used for debug reasons: */
void
hex_dump(unsigned char *a, size_t len)
{
	for (size_t i = 0; i < len; i++)
		printf("%02x", a[i]);
}

void
do_xor(unsigned char *xored, const unsigned char *left,
    const unsigned char *right, const unsigned int bytes)
{
	for (unsigned int i = 0; i < bytes; i++)
		xored[i] = XOR(left[i], right[i]);
}

void
toByte(uint8_t *buf, const uint32_t x, const uint32_t y)
{
	/*
	 * If x and y are non-negative integers, we define Z = toByte(x,y) to be
	 * the y-byte string containing the binary representation of x in big
	 * endian byte-order.
	 */

	buf[y-1] = x & 0xFF;
	if ( y > 1 )
		buf[y-2] = (x >> 8) & 0xFF;
	if ( y > 2)
		buf[y-3] = (x >> 16) & 0xFF;
	if ( y > 3 )
		buf[y-4] = (x >> 24) & 0xFF;
	if ( y > 4 )
		memset(&buf[0], 0, y * sizeof(*buf) - 4);
}

/* This is algorithm 1: base_w(X,w) */
void
base_w(const unsigned char *X, const unsigned int w, const unsigned int len_X,
    int *bas)
{
	unsigned int in = 0, out = 0, total = 0, bits = 0, consumed = 0;

	for (consumed = 0; consumed < 8 * len_X; consumed += (unsigned int)log2(w)) {
		if (bits == 0) {
			total = X[in];
			in++;
			bits += 8;
		}
		bits -= (int)log2(w);
		bas[out] = (total >> bits) & (w - 1);
		out++;
	}
}

/* This is algorithm 2: the chaining function */
int
chain(unsigned char *chained, const unsigned char *X, const unsigned int i,
    const unsigned int s, const wots_param *para, unsigned char *seed,
    unsigned char adrs[16])
{
	/* returns 1 if i + s is too large (greater than w - 1) */
	unsigned int bytes = para->m;
	if (s == 0) {
		memcpy(chained, X, bytes * sizeof(*chained));
		return 0;
	}
	if ((i + s) > ((para->w) - 1))
		return 1;

	unsigned char tmp[bytes];
	memset(tmp, 0, bytes * sizeof(*tmp));
	chain(tmp, X, i, s - 1, para, seed, adrs);
	setHashAddress(adrs, i + s - 1);
	setKeyBit(adrs, 0);
	unsigned char bitmask[bytes];
	memset(bitmask, 0, bytes * sizeof(*bitmask));
	G(bitmask, bytes, seed, adrs, bytes);
	setKeyBit(adrs, 1);
	unsigned char key[bytes];
	memset(key, 0, bytes * sizeof(*key));
	G(key, bytes, seed, adrs, bytes);
	unsigned char xored[bytes];
	memset(xored, 0, bytes * sizeof(*xored));
	do_xor(xored, tmp, bitmask, bytes);
	F(chained, xored, key, bytes);

	return 0;
}

/*
 * This is algorithm 3: generating the WOTS+ private key
 * Take care of a suitable random source (uniformely at random).
 */
int
genWotsPrivKey(const wots_param *para, unsigned char *sk)
{
	/* returns 1 if call to RAND_bytes failed */
	unsigned int bytes = para->m, len = para->len;
	unsigned char key[bytes];
	int check = 0;

	for (unsigned int i = 0; i < len; i++) {
		check = RAND_bytes(key, bytes);
		if (check != 1)
			return 1;
		memcpy(&sk[i*bytes], key, bytes * sizeof(*sk));
	}

	return 0;
}

/* This is algorithm 4: */
void
genWotsPubKey(unsigned char *pk, const wots_param *para,
    const unsigned char *sk, unsigned char *seed, unsigned char adrs[16])
{
	unsigned int len = para->len, w = para->w;

	for (unsigned int i = 0; i < len; i++) {
		setChainAddress(adrs, i);
		chain(&(pk[i*(para->m)]), &(sk[i*(para->m)]), 0, w-1, para,
		    seed, adrs); /* Writing to pk[i] */
	}
}

/* This is algorithm 5: */
void
genWotsSig(const unsigned char *sk, const wots_param *para,
    const unsigned char *M, unsigned char *sig, unsigned char *seed,
    unsigned char adrs[16])
{
	unsigned int i = 0, w = para->w;
	int csum = 0;
	unsigned int len = para->len, len_1 = para->len_1, len_2 = para->len_2;
	int tmp_msg[len_1];
	memset(tmp_msg, 0, len_1 * sizeof(*tmp_msg));
	int msg[len];
	memset(msg, 0, len * sizeof(*msg));
	base_w(M, w, para->m, tmp_msg);
	unsigned int len_2_bytes = 0;
	for (i = 0; i < len_1; i = i + 1)
		csum = csum + (int)w - 1 - tmp_msg[i];
	csum = csum << (8 - (int)((int)(len_2 * log2(w)) % 8));
	len_2_bytes = ((len_2 * (int)log2(w)) + 7) / 8;
	unsigned char byte[len_2_bytes];
	memset(byte, 0, len_2_bytes * sizeof(*byte));
	toByte(byte, csum, len_2_bytes);
	int tmp[len_2];
	for (unsigned int q = 0 ; q < len_2; q++)
		tmp[q] = 0;
	base_w(byte, w, len_2_bytes, tmp);
	/* Pseudo: msg = msg || tmp; */
	concat_int(msg, tmp_msg, len_1, tmp, len_2);
	for (i = 0; i < len; i++) {
		setChainAddress(adrs, i);
		chain(&sig[i*para->m], &sk[i*para->m], 0,
		    (unsigned int)msg[i], para, seed, adrs);
	}
}

/*
 * This is algorithm 6:
 * Please note that we have a 'temporary' WOTS+ public key tmp_pk
 * which is yet to be compared to the actual WOTS+ public key. 
 * This is done by XMSS.
 */
void
wotsPubKeyFromSig(const wots_param *para, const unsigned char *M,
    const unsigned char *sig, unsigned char *tmp_pk, unsigned char *seed,
    unsigned char adrs[16])
{
	unsigned int i = 0, w = para->w;
	int csum = 0;
	unsigned int len_2_bytes = 0;
	unsigned int len = para->len, len_1 = para->len_1, len_2 = para->len_2;
	
	/* Transform to base_w representation */
	int tmp_msg[len_1];
	memset(tmp_msg, 0, len_1 * sizeof(*tmp_msg));
	int msg[len];
	memset(msg, 0, len * sizeof(*msg));
	base_w(M, w, para->m, tmp_msg);

	/* Compute the checksum */
	for (i = 0; i < len_1; i++)
		csum = csum + (int)w - 1 - tmp_msg[i];

	/* Convert csum to a base_w representation */
	csum = csum << ( 8 - (int)((int)(len_2 * log2(w)) % 8));
	len_2_bytes = ((len_2 * (int)log2(w)) + 7) / 8;

	/* Build the actual message */
	unsigned char byte[len_2_bytes];
	memset(byte, 0, len_2_bytes * sizeof(*byte));
	toByte(byte, csum, len_2_bytes);
	int tmp[len_2];
	for (unsigned int q = 0 ; q < len_2; q++)
		tmp[q] = 0;
	base_w(byte, w, len_2_bytes, tmp);

	/* Pseudo: msg = msg || tmp; */
	concat_int(msg, tmp_msg, len_1, tmp, len_2);
	for (i = 0; i < len; i++) {
		setChainAddress(adrs, i);
		chain(&tmp_pk[i * para->m], &sig[i * para->m],
		    (unsigned int)msg[i], w - 1 - (unsigned int)msg[i],
		    para, seed, adrs);
	}
}

unsigned int
calcLen1(const unsigned int w, const unsigned int m)
{
	unsigned int len_1 =
	    (unsigned int)ceil(8.0 * (double)m / (int)log2((double)w));
	return len_1;
}

unsigned int
calcLen2(const unsigned int w, const unsigned int len_1)
{
	unsigned int len_2 = (unsigned int)floor((int)log2((double)len_1 *
	    ((double)w - 1.0)) / (int)log2((double)w)) + 1;
	return len_2;
}

/* Most trivial function, but let's go for readability: */
unsigned int
calcLen(const unsigned int len_1, const unsigned int len_2)
{
	return len_1 + len_2;
}

int
wots_param_init(wots_param *para, const unsigned int m, const unsigned int n,
    const unsigned int w)
{
	if (!((m == 32 || m == 64) && (w == 4 || w == 16)))
		return 1;

	para->m = m;
	para->n = n;
	para->w = w;
	para->len_1 = (unsigned int)ceil(8.0 * (double)para->m /
	    log2((double)para->w));
	para->len_2 = (unsigned int)floor(log2((double)para->len_1 *
	    ((double)para->w - 1.0)) / log2((double)para->w)) + 1;
	para->len = para->len_1 + para->len_2;

	return 0;
}

void
concat_int(int *concat, const int *left, const unsigned int bytes_left,
    const int *right, const unsigned int bytes_right)
{
	memcpy(concat, left, bytes_left * sizeof(int));
	memcpy(concat + bytes_left, right, bytes_right * sizeof(int));
}
