/*
 * Copyright (c) 2016, Denis Butin and Stefan-Lukas Gazdag
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "xmss_draft.h"

/*
 * This is algorithm 7: RAND_HASH
 * It's used to show randomized hashing in the draft.
 * But it's also useful as separate function here.
 */
void
rand_hash(unsigned char *hash, unsigned char *left, unsigned char *right,
    unsigned char *seed, unsigned char adrs[16], const unsigned int bytes)
{
	unsigned char bm_0[bytes], bm_1[bytes], key[bytes];
	memset(bm_0, 0, bytes * sizeof(unsigned char));
	memset(bm_1, 0, bytes * sizeof(unsigned char));
	memset(key, 0, bytes * sizeof(unsigned char));
	setKeyBit(adrs, 0);
	setBlockBit(adrs, 0);
	G(bm_0, bytes, seed, adrs, bytes);
	setBlockBit(adrs, 1);
	G(bm_1, bytes, seed, adrs, bytes);
	setKeyBit(adrs, 1);
	setBlockBit(adrs, 0);
	G(key, bytes, seed, adrs, bytes);
	unsigned char xor_left[bytes], xor_right[bytes];
	memset(xor_left, 0, bytes * sizeof(unsigned char));
	memset(xor_right, 0, bytes * sizeof(unsigned char));
	do_xor(xor_left, left, bm_0, bytes);
	do_xor(xor_right, right, bm_1, bytes);
	unsigned char concat[bytes+bytes];
	memset(concat, 0, (bytes + bytes) * sizeof(unsigned char));
	concat_bytes(concat, xor_left, bytes, xor_right, bytes);
	H(hash, concat, key, bytes);
	/* Pseudo: H(key, (XOR(left, bm_0)) || (XOR(right, bm_1))); */
}

/* This is algorithm 8: ltree */
void
ltree (unsigned char *pk, unsigned char *node, const wots_param *para,
    unsigned char *seed, unsigned char adrs[16])
{
	unsigned int m = para->m, len_tmp = para->len, i = 0, treeHeight = 0;
	setTreeHeight(adrs, 0);
	while (len_tmp > 1) {
		for (i = 0; i < (unsigned int)floor(len_tmp / 2.0); i++) {
			setTreeIndex(adrs, i);
			/* setting pk[i] */
			rand_hash(&pk[i * m], &pk[2 * i * m],
			    &pk[(2 * i + 1) * m], seed, adrs, m);
		}
		if (len_tmp % 2 == 1) {
			memcpy(&pk[((int)floor(len_tmp / 2.0)) * m],
			    &pk[(len_tmp - 1) * m], m * sizeof(unsigned char));
		}
		len_tmp = (unsigned int)ceil(len_tmp / 2.0);
		treeHeight++;
		setTreeHeight(adrs, treeHeight);
	}
	memcpy(node, &pk[0], m * sizeof(unsigned char));
}

/* This is algorithm 9: */
int
treeHash (unsigned char *hash, xmss_sk *SK, const unsigned int s,
    const unsigned int t, unsigned char *seed, unsigned char adrs[16])
{
	/*
	 * returns 1 if s is not a multiple of 2^t
	 * returns 2 if next access to SK->wots_keys would be out of bounds
	 */
	unsigned int m = SK->bytes;
	xmss_node stack[t + 1];
	wots_param par = SK->para;
	for (unsigned int i = 0 ; i < (t + 1); i++) {
		memset(stack[i].node, 0, m * sizeof(unsigned char));
		stack[i].height = 999;
	}
	int nodenumber = 0, topnode = -1;
	unsigned char pk[par.len * m], node[m], tmp_node[m];
	memset(pk, 0, par.len * m * sizeof(unsigned char));
	memset(node, 0, m * sizeof(unsigned char));
	memset(tmp_node, 0, m * sizeof(unsigned char));
	unsigned int height = 0, index = 0;
	/* This is just a single Wots secret key: */
	unsigned char sk[par.len * m];
	memset(sk, 0, par.len * m * sizeof(unsigned char));
	unsigned int sum = 0;
	/*
	 * Specifically for XMSS^MT:
	 * Don't generate WOTS keys if already done, so check if empty
	 */
	for (unsigned int i = 0; i < ((1 << t) * par.len * m); ++i)
		sum |= SK->wots_keys[i];

	if ((s == 0) && (t == SK->h) && (sum == 0)) {
		for (unsigned int num = 0; num < (unsigned int)(1<<t); num++) {
			genWotsPrivKey(&SK->para, sk);
			memcpy(&SK->wots_keys[num * par.len * m], sk,
			    par.len * m * sizeof(unsigned char));
		}
	}

	if (s % ( 1 << t ) != 0)
		return 1;

	for (unsigned int i = 0; i < (unsigned int)(1 << t); i++) {
		setOTSBit(adrs, 1);
		setOTSAddress(adrs, s + i);
		/* getWOTS_SK(SK, sk, s+i); */

		if ((s + i) > (unsigned int)((1 << SK->h) - 1))
			return 2;

		memcpy(sk, &SK->wots_keys[(s + i) * par.len * m],
		    par.len * m * sizeof(unsigned char));

		genWotsPubKey(pk, &SK->para, sk, seed, adrs);
		setOTSBit(adrs, 0);
		setLTreeBit(adrs, 1);
		setLTreeAddress(adrs, (s + i));
		ltree(pk, node, &SK->para, seed, adrs);
		setLTreeBit(adrs, 0);
		setTreeHeight(adrs, 0);
		height = 0;
		setTreeIndex(adrs, (s + i));
		index = i + s;
		/* Not related to draft: Save all nodes to SK */
		if ((s == 0) && (t == SK->h)) {
			memcpy(SK->nodes[nodenumber].node, node,
			    m * sizeof(unsigned char));
			SK->nodes[nodenumber].height = height;
			nodenumber++;
		}

		while (height == stack[topnode].height && topnode > -1) {
			getTreeIndex(adrs, &index);
			setTreeIndex(adrs, ((index - 1) / 2));
			rand_hash(tmp_node, stack[topnode].node, node,
			    seed, adrs, m);
			memset(stack[topnode].node, 0,
			    m * sizeof(unsigned char));
			stack[topnode].height = 0;
			topnode--;
			memcpy(node, tmp_node,
			    m * sizeof(unsigned char));
			getTreeHeight(adrs, &height);
			setTreeHeight(adrs, ++height);
			if ((s == 0) && (t == SK->h)) {
				memcpy(SK->nodes[nodenumber].node, node,
				    m * sizeof(unsigned char));
				SK->nodes[nodenumber].height = height;
				nodenumber++;
			}
		}
		topnode++;
		memcpy(stack[topnode].node, node, m * sizeof(unsigned char));
		stack[topnode].height = height;
	}
	memcpy(hash, stack[topnode].node, SK->bytes * sizeof(unsigned char));
	return 0;
}

/* Initialize the XMSS secret key SK */
int
xmss_sk_init(xmss_sk *SK, const wots_param *para, const unsigned int h)
{
	/*
	 * returns 1 if a problem with calloc occurred
	 * returns 2 if height h is less than 2
	 * returns 3 if call to RAND_bytes for SK_PRF failed
	 */
	SK->idx = 0;
	SK->bytes = para->m;
	SK->para = *para;
	if (h < 2)
		return 2;
	SK->h = h;
	if ((SK->SK_PRF = calloc(para->m, sizeof(unsigned char))) == NULL)
		return 1;
	int rcprf = RAND_bytes(SK->SK_PRF, para->m);
 	if (rcprf != 1)
		return 3;
	if ((SK->nodes = calloc(((2 << h)) - 1, sizeof(xmss_node))) == NULL)
		return 1;
	if ((SK->wots_keys = calloc((para->len * para->m * (1 << h)),
	    sizeof(unsigned char))) == NULL)
		return 1;
	return 0;
}

void
xmss_sk_free(xmss_sk *SK)
{
	free(SK->SK_PRF);
	free(SK->nodes);
	free(SK->wots_keys);
}

int
xmss_sig_init(xmss_sig *Sig, const wots_param *para, const unsigned int h)
{
	/* returns 1 if a problem with calloc occurred */
	Sig->idx_sig = 0;
	if ((Sig->randomness = calloc(para->m, sizeof(unsigned char))) == NULL)
		return 1;
	if ((Sig->sig_ots = calloc(para->len * para->m,
	    sizeof(unsigned char))) == NULL)
		return 1;
	if ((Sig->auth = calloc(h * para->m, sizeof(unsigned char))) == NULL)
		return 1;
	return 0;
}

void
xmss_sig_free(xmss_sig *Sig)
{
	free(Sig->randomness);
	free(Sig->sig_ots);
	free(Sig->auth);
}

int
xmss_pk_init(xmss_pk *PK, const unsigned int bytes)
{
	/* returns 1 if a problem with calloc occurred */
	if ((PK->root = calloc(bytes, sizeof(unsigned char))) == NULL)
		return 1;
	if ((PK->seed = calloc(bytes, sizeof(unsigned char))) == NULL)
		return 1;
	return 0;
}

void
xmss_pk_free(xmss_pk *PK)
{
	free(PK->root);
	free(PK->seed);
}

/* This is algorithm 10: */
int
xmss_genpk (unsigned char *pk, xmss_sk *SK, unsigned int h)
{
	/*
	 * returns 1 if initialization of xmss_pk failed
	 * returns 2 if call to RAND_bytes failed
	 */
	xmss_pk PK;
	if ((xmss_pk_init(&PK, SK->bytes)) == 1)
		return 1;
	unsigned int check = 0;
	check = RAND_bytes(PK.seed, SK->bytes);
	if (check != 1)
		return 2;
	unsigned char adrs[16];
	memset(adrs, 0, 16 * sizeof(unsigned char)); /* toByte(adrs, 0, 16) */
	treeHash(PK.root, SK, 0, h, PK.seed, adrs);
	xmssPKtoString(pk, &PK, SK->bytes);
	xmss_pk_free(&PK);
	return 0;
}

/*
 * This is an adaption for the non-numbered algorithm for the calculation
 * of the authentication path:
 */
int
buildAuth (xmss_sk *SK, unsigned char *auth_path, const uint32_t i)
{
	/* returns 1 if no suitable node could be found */
	unsigned int m = SK->bytes, h = SK->h, k = 0, occurrence = 0;
	unsigned char auth[h * m];
	uint32_t idx = i;
	/*
	 * We have SK->nodes, which is an array containing all nodes as
	 * xmss_node which is a struct containing the n-byte node string and
	 * the height
	 */
	memset(auth, 0, h * m * sizeof(unsigned char));
	for (unsigned int j = 0; j < h; j++) {
		/*
		 * KISS: A simple but inefficient implementation of buildAuth.
		 * Should be replaced by a more efficient variant in actual
		 * applications.
		 */
		if ((idx % 2) == 0) {
			while (occurrence != (idx + 2)) {
				if (SK->nodes[k].height == j)
					occurrence++;
				k++;
				if (k == (unsigned int)((1 << (h + 1)) - 1))
					return 1;
			}
			k--;
			while (SK->nodes[k].height != j)
				k--;
			memcpy(&auth[j * m], SK->nodes[k].node,
			    m * sizeof(unsigned char));
			idx = idx / 2;
		}
		else
		{
			while (occurrence != idx) {
				if (SK->nodes[k].height == j)
					occurrence++;
				k++;
				if (k == (unsigned int)((1 << (h + 1)) - 1))
					return 1;
			}
			k--;
			memcpy(&auth[j * m], SK->nodes[k].node,
			    m * sizeof(unsigned char));
			idx = (idx - 1) / 2;
		}
		k = 0;
		occurrence = 0;
	}
	memcpy(auth_path, &auth, h * m * sizeof(unsigned char));
	return 0;
}

/* This is a getter for receiving the WOTS sk */
int
getWOTS_SK(const xmss_sk *SK, unsigned char *key, const unsigned int index)
{
	/*
	 * returns 1 if the access to SK->wots_keys via index
	 * would be out of bounds
	 */
	if(index > (unsigned int)((1 << SK->h) - 1))
		return 1;
	memcpy(key, &SK->wots_keys[index * SK->bytes * SK->para.len],
	    (SK->bytes * SK->para.len) * sizeof(unsigned char));
	return 0;
}

void
getSK_PRF(const xmss_sk *SK, unsigned char *sk_prf)
{
	memcpy(sk_prf, SK->SK_PRF, SK->bytes * sizeof(unsigned char));
}

/* This is algorithm 11: */
int
xmss_signing(xmss_sk *SK, unsigned char *msg, const unsigned int msglen,
    unsigned char *sig, unsigned char *seed, unsigned char adrs[16])
{
	/*
	 * returns 1 if the initialization of xmss_sig failed
	 * returns 2 if the index idx_sig is too large for the tree
	 * returns 3 if the access to SK->wots_keys would be out of bounds
	 */
	wots_param para = SK->para;
	xmss_sig Sig;
	if ((xmss_sig_init(&Sig, &para, SK->h)) == 1)
		return 1;
	xmssStringToSig(&Sig, sig, para.m, SK->bytes, SK->h, para.w);
	uint32_t idx_sig = SK->idx;
	unsigned int h = SK->h;
	if (idx_sig >= (uint32_t)(1 << h))
		return 2;
	else if ((idx_sig + 1) == (uint32_t)(1 << h))
	{
		/*
		 * One may want to be informed that this is the last message to
		 * be signed with this tree
		 */
	}
	unsigned int bytes = SK->bytes, len = SK->para.len;
	unsigned char auth[h * bytes];
	memset(auth, 0, h * bytes * sizeof(unsigned char));
	buildAuth(SK, auth, idx_sig);
	unsigned char r[bytes], sk_prf[bytes];
	memset(r, 0, bytes * sizeof(unsigned char));
	memset(sk_prf, 0, bytes * sizeof(unsigned char));
	getSK_PRF(SK, sk_prf);
	PRFm(r, bytes, msg, msglen, sk_prf, bytes); /* byte[m] */
	unsigned char msg_tmp[bytes];
	memset(msg_tmp, 0, bytes * sizeof(unsigned char));
	unsigned char input_tmp[2 * bytes];
	memset(input_tmp, 0, 2 * bytes * sizeof(unsigned char));
	unsigned char tobyte_tmp[bytes];
	memset(tobyte_tmp, 0, bytes * sizeof(unsigned char));
	toByte(tobyte_tmp, idx_sig, bytes);
	concat_bytes(input_tmp, tobyte_tmp, bytes, r, bytes);
	Hm(msg_tmp, msg, msglen, input_tmp, bytes); /* byte[m] */
	setOTSBit(adrs, 1);
	setOTSAddress(adrs, idx_sig);
	unsigned char sig_ots[len * bytes];
	memset(sig_ots, 0, len * bytes * sizeof(unsigned char));
	if (idx_sig > (uint32_t)((1 << h) - 1))
		return 3;
	genWotsSig(&SK->wots_keys[idx_sig * bytes * len], &SK->para, msg_tmp,
	    sig_ots, seed, adrs);
	/* Pseudo: Sig = (idx_sig || r || sig_ots || auth); */
	memcpy(Sig.randomness, r, bytes * sizeof(unsigned char));
	memcpy(Sig.sig_ots, sig_ots, len * bytes * sizeof(unsigned char));
	memcpy(Sig.auth, auth, h * bytes * sizeof(unsigned char));
	Sig.idx_sig = SK->idx;
	SK->idx = idx_sig + 1;
	xmssSigToString(sig, &Sig, para.m, SK->bytes, SK->h, para.w);
	xmss_sig_free(&Sig);
	return 0;
}

/* This is algorithm 12: */
int
xmss_rootFromSig(xmss_sig *Sig, unsigned char *msg, const unsigned int msglen,
    unsigned char *node, unsigned char *seed, const unsigned int bytes,
    const unsigned int h, unsigned char adrs[16])
{
	/*
	 * returns 1 if the index idx_sig is too large for the tree
	 * Note: we actually want to get h (the height of the tree)
	 * from the OID. This is just a workaround
	 * If you actually get it from Sig: bytes 4 to m+3
	 */
	unsigned char r[bytes];
	memset(r, 0, bytes * sizeof(unsigned char));
	/* If you actually get it from Sig: the last h-1 * n bytes */
	unsigned char auth[h * bytes];
	memset(auth, 0, h * bytes * sizeof(unsigned char));
	memcpy(r, Sig->randomness, bytes * sizeof(unsigned char));
	memcpy(auth, Sig->auth, h * bytes * sizeof(unsigned char));
	/* If you actually get it from Sig: bytes 0 - 3 */
	uint32_t idx_sig = Sig->idx_sig;
	if (idx_sig >= (uint32_t)(1 << h))
		return 1;
	wots_param para;
	para.m = bytes;
	para.w = 16;
	para.len_1 = calcLen1(para.w, para.m);
	para.len_2 = calcLen2(para.w, para.len_1);
	para.len = calcLen(para.len_1, para.len_2);
	unsigned char sig_ots[para.len * bytes];
	memset(sig_ots, 0, para.len * bytes * sizeof(unsigned char));
	memcpy(sig_ots, Sig->sig_ots, para.len * bytes * sizeof(unsigned char));
	unsigned char msg_tmp[bytes];
	memset(msg_tmp, 0, bytes * sizeof(unsigned char));
	unsigned char input_tmp[2 * bytes];
	memset(input_tmp, 0, 2 * bytes * sizeof(unsigned char));
	unsigned char tobyte_tmp[bytes];
	memset(tobyte_tmp, 0, bytes * sizeof(unsigned char));
	toByte(tobyte_tmp, idx_sig, bytes);
	concat_bytes(input_tmp, tobyte_tmp, bytes, r, bytes);
	Hm(msg_tmp, msg, msglen, input_tmp, bytes); /* byte[m] */
	setOTSBit(adrs, 1);
	setOTSAddress(adrs, idx_sig);
	unsigned char pk_ots[para.len * bytes];
	memset(pk_ots, 0, para.len * bytes * sizeof(unsigned char));
	wotsPubKeyFromSig(&para, msg_tmp, sig_ots, pk_ots, seed, adrs);
	setOTSBit(adrs, 0);
	setLTreeBit(adrs, 1);
	setLTreeAddress(adrs, idx_sig);
	unsigned char node0[para.m], node1[para.m];
	memset(node0, 0, para.m * sizeof(unsigned char));
	memset(node1, 0, para.m * sizeof(unsigned char));
	ltree(pk_ots, node0, &para, seed, adrs);
	setLTreeBit(adrs, 0);
	setTreeIndex(adrs, idx_sig);
	unsigned int k = 0, index = 0;
	for (k = 0; k < h; k++) {
		setTreeHeight(adrs, k);
		if (((int)floor((double)((idx_sig / (1 << k)) % 2))) == 0) {
			getTreeIndex(adrs, &index);
			setTreeIndex(adrs, (index >> 1));
			rand_hash(node1, node0, &auth[k * bytes], seed, adrs,
			    bytes);
		}
		else
		{
			getTreeIndex(adrs, &index);
			setTreeIndex(adrs, ((index - 1) >> 1));
			rand_hash(node1, &auth[k * bytes], node0, seed, adrs,
			    bytes);
		}
		memcpy(node0, node1, para.m * sizeof(unsigned char));
	}
	memcpy(node, node0, para.m * sizeof(unsigned char));
	return 0;
}

/* This is algorithm 13: */
int
xmss_verification(unsigned char *pk_all, unsigned char *sig, unsigned char *msg,
    const unsigned int msglen, const wots_param *para, const unsigned int h)
{
	/*
	 * returns 0 if verification was successful
	 * returns 1 if verification was NOT successful
	 * returns 2 if index is too large for the tree
	 */
	xmss_pk PK;
	xmss_pk_init(&PK, para->m);
	xmssStringToPK(&PK, pk_all, para->m);
	xmss_sig Sig;
	xmss_sig_init(&Sig, para, h);
	unsigned int m = para->m, w = para->w;
	xmssStringToSig(&Sig, sig, m, m, h, w);
	if (Sig.idx_sig >= (uint32_t)(1 << h))
		return 2;
	unsigned char node[m], pk[m], seed[m];
	memset(node, 0, m * sizeof(unsigned char));
	memset(pk, 0, m * sizeof(unsigned char));
	memcpy(pk, PK.root, m * sizeof(unsigned char));
	memset(seed, 0, m * sizeof(unsigned char));
	memcpy(seed, PK.seed, m * sizeof(unsigned char));
	unsigned char adrs[16] = { 0 };
	xmss_rootFromSig(&Sig, msg, msglen, node, seed, m, h, adrs);
	xmss_sig_free(&Sig);
	xmss_pk_free(&PK);
	if (memcmp(node, pk, m) == 0)
		return 0;
	return 1; /* 0 if verification successful, else 1 */
}

/*
 * Conversion from a xmss_sig to an unsigned char *.
 * concat_bytes requires the length in bytes of the strings to concatenate,
 * and these lengths in turn depend on m, n, len and h.
 * We don't need len as an explicit parameter however, since it only requires w
 * in addition to the rest and we assume w = 16.
 */
void
xmssSigToString(unsigned char *concat, xmss_sig *Sig, const int m, const int n,
    const int h, const int w)
{
	unsigned int len1 = calcLen1(w, m);
	unsigned int len = calcLen(len1, calcLen2(w, len1));
	unsigned char idx_sig_char[4] = { 0 };
	uint32_t idx_sig = Sig->idx_sig;
	idx_sig_char[0] = (unsigned char)(idx_sig >> 24) & 0xFF;
	idx_sig_char[1] = (unsigned char)(idx_sig >> 16) & 0xFF;
	idx_sig_char[2] = (unsigned char)(idx_sig >> 8) & 0xFF;
	idx_sig_char[3] = (unsigned char)(idx_sig) & 0xFF;
	/* concat idx_sig and randomness */
	concat_bytes(concat, idx_sig_char, 4, Sig->randomness, m);
	/* concat the above and sig_ots */
	concat_bytes(concat, concat, 4 + m, Sig->sig_ots, n * len);
	/* concat the above and auth */
	concat_bytes(concat, concat, (n * len) + 4 + m, Sig->auth, h * n);
}

/* Conversion from a xmss_pk to an unsigned char * */
void
xmssPKtoString(unsigned char *concat, const xmss_pk *PK, const int n)
{
	/* concat root and seed */
	concat_bytes(concat, PK->root, n, PK->seed, n);
}

void
xmssStringToSig(xmss_sig *Sig, unsigned char *string, const int m, const int n,
    const int h, const int w)
{
	unsigned int len1 = calcLen1(w, m);
	unsigned int len = calcLen(len1, calcLen2(w, len1));
	uint32_t idx_sig = 0;
	idx_sig = (uint32_t)string[0] << 24 | (uint32_t)string[1] << 16 |
	    (uint32_t)string[2] << 8 | (uint32_t)string[3];
	Sig->idx_sig = idx_sig;
	memcpy(Sig->randomness, &string[4], m * sizeof(unsigned char));
	memcpy(Sig->sig_ots, &string[4 + m], len * n * sizeof(unsigned char));
	memcpy(Sig->auth, &string[4 + m + (len * n)],
	    h * n * sizeof(unsigned char));
}

void
xmssStringToPK(xmss_pk *PK, unsigned char *string, const int n)
{
	memcpy(&PK->root[0], &string[0], n * sizeof(unsigned char));
	memcpy(&PK->seed[0], &string[n], n * sizeof(unsigned char));
}

